// Copyright 2024 Pixar
//
// Licensed under the terms set forth in the LICENSE.txt file available at
// https://openusd.org/license.
//
// Modified by Jeremy Retailleau.

#include "./raii.h"
#include <pxr/tf/instantiateStacked.h>

namespace pxr {


TF_INSTANTIATE_DEFINED_STACKED(TsAntiRegressionAuthoringSelector);

TF_INSTANTIATE_DEFINED_STACKED(TsEditBehaviorBlock);


}  // namespace pxr
